# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides a set of functions to handle resampling.

*License*:
    Copyright (c) 2018, SDSS-IV/MaNGA Pipeline Group
        Licensed under BSD 3-clause license - see LICENSE.rst

*Revision history*:
    | **30 Aug 2018**: Moved here from old instrument.py module by K.
        Westfall (KBW).

.. include common links, assuming primary doc root is up one directory
.. include:: ../links.rst
"""

import warnings
import numpy
from scipy import interpolate
import astropy.constants

from pypeit.core import moment

def spectral_coordinate_step(wave, log=False, base=10.0):
    """
    Return the sampling step for the input wavelength vector.

    If the sampling is logarithmic, return the change in the logarithm
    of the wavelength; otherwise, return the linear step in angstroms.

    Args: 
        wave (numpy.ndarray): Wavelength coordinates of each spectral
            channel in angstroms.
        log (bool): (**Optional**) Input spectrum has been sampled
            geometrically.
        base (float): (**Optional**) If sampled geometrically, the
            sampling is done using a logarithm with this base.  For
            natural logarithm, use numpy.exp(1).

    Returns:
        float: Spectral sampling step in either angstroms (log=False) or
        the step in log(angstroms).
    """
    dw = numpy.diff(numpy.log(wave))/numpy.log(base) if log else numpy.diff(wave)
    if numpy.any( numpy.absolute(numpy.diff(dw)) > 100*numpy.finfo(dw.dtype).eps):
        raise ValueError('Wavelength vector is not uniformly sampled to numerical accuracy.')
    return numpy.mean(dw)


def spectrum_velocity_scale(wave):
    """
    Determine the velocity sampling of an input wavelength vector when log sampled
    
    .. note::
        The wavelength vector is assumed to be geometrically sampled!
        However, the input units expected to be in angstroms, not, e.g.,
        log(angstrom).

    Args: 
        wave (numpy.ndarray): Wavelength coordinates of each spectral
            channel in angstroms.  It is expected that the spectrum has
            been sampled geometrically

    Returns:
        float: Velocity scale of the spectrum in km/s.

    """
    return astropy.constants.c.to('km/s').value*spectral_coordinate_step(wave, log=True,
                                                                         base=numpy.exp(1.))


def angstroms_per_pixel(wave, log=False, base=10.0, regular=True):
    """
    Return a vector with the angstroms per pixel at each channel.

    When `regular=True`, the function assumes that the wavelengths are
    either sampled linearly or geometrically.  Otherwise, it calculates
    the size of each pixel as the difference between the wavelength
    coordinates.  The first and last pixels are assumed to have a width
    as determined by assuming the coordinate is at its center.

    Args:
        wave (`numpy.ndarray`):
            (Geometric) centers of the spectrum pixels in angstroms.
        log (`numpy.ndarray`, optional):
            The vector is geometrically sampled.
        base (:obj:`float`, optional):
            Base of the logarithm used in the geometric sampling.
        regular (:obj:`bool`, optional):
            The vector is regularly sampled.

    Returns:
        numpy.ndarray: The angstroms per pixel.
    """
    if regular:
        ang_per_pix = spectral_coordinate_step(wave, log=log, base=base)
        if log:
            ang_per_pix *= wave*numpy.log(base)
    else:
        ang_per_pix = numpy.diff([(3*wave[0]-wave[1])/2] 
                                    + ((wave[1:] + wave[:-1])/2).tolist()
                                    + [(3*wave[-1]-wave[-2])/2])
    return ang_per_pix


def _pixel_centers(xlim, npix, log=False, base=10.0):
    """
    Determine the centers of pixels in a linearly or geometrically
    sampled vector given first, last and number of pixels

    Args:
        xlim (numpy.ndarray) : (Geometric) Centers of the first and last
            pixel in the vector.
        npix (int) : Number of pixels in the vector.
        log (bool) : (**Optional**) The input range is (to be)
            logarithmically sampled.
        base (float) : (**Optional**) The base of the logarithmic
            sampling.  The default is 10.0; use numpy.exp(1.) for the
            natural logarithm.

    Returns:
        numpy.ndarray, float: A vector with the npix centres of the
        pixels and the sampling rate.  If logarithmically binned, the
        sampling is the step in :math`\log x`.
    """
    if log:
        logRange = numpy.log(xlim)/numpy.log(base)
        dlogx = numpy.diff(logRange)/(npix-1.)
        centers = numpy.power(base, numpy.linspace(*(logRange/dlogx), num=npix)*dlogx)
        return centers, dlogx
    dx = numpy.diff(xlim)/(npix-1.)
    centers = numpy.linspace(*(xlim/dx), num=npix)*dx
    return centers, dx


def _pixel_borders(xlim, npix, log=False, base=10.0):
    """
    Determine the borders of the pixels in a vector given the first, last and 
    number of pixels

    Args:
        xlim (numpy.ndarray) : (Geometric) Centers of the first and last
            pixel in the vector.
        npix (int) : Number of pixels in the vector.
        log (bool) : (**Optional**) The input range is (to be)
            logarithmically sampled.
        base (float) : (**Optional**) The base of the logarithmic
            sampling.  The default is 10.0; use numpy.exp(1.) for the
            natural logarithm.

    Returns:
        numpy.ndarray, float: A vector with the (npix+1) borders of the
        pixels and the sampling rate.  If logarithmically binned, the
        sampling is the step in :math`\log x`.
    """
    if log:
        logRange = numpy.log(xlim)/numpy.log(base)
        dlogx = numpy.diff(logRange)/(npix-1.)
        borders = numpy.power(base, numpy.linspace(*(logRange/dlogx + [-0.5, 0.5]),
                                                   num=npix+1)*dlogx)
        return borders, dlogx
    dx = numpy.diff(xlim)/(npix-1.)
    borders = numpy.linspace(*(xlim/dx + numpy.array([-0.5, 0.5])), num=npix+1)*dx
    return borders, dx


def resample_vector_npix(outRange=None, dx=None, log=False, base=10.0, default=None):
    """
    Determine the number of pixels needed to resample a vector given first, last pixel and dx

    Args:
        outRange (list or numpy.ndarray) : Two-element array with the
            starting and ending x coordinate of the pixel centers to
            divide into pixels of a given width.  If *log* is True, this
            must still be the linear value of the x coordinate, not
            log(x)!.
        dx (float) : Linear or logarithmic pixel width.
        log (bool) : Flag that the range should be logarithmically
            binned.
        base (float) : Base for the logarithm
        default (int) : Default number of pixels to use.  The default is
            returned if either *outRange* or *dx* are not provided.

    Returns:
        int, numpy.ndarray: Returns two objects: The number of pixels to
        cover *outRange* with pixels of width *dx* and the adjusted
        range such that number of pixels of size dx is the exact integer.

    Raises:
        ValueError: Raised if the range is not a two-element vector
    """
    # If the range or sampling are not provided, the number of pixels is
    # already set
    if outRange is None or dx is None:
        return default, outRange
    if len(outRange) != 2:
        raise ValueError('Output range must be a 2-element vector.')

    _outRange = numpy.atleast_1d(outRange).copy()
    npix = int( numpy.diff(numpy.log(_outRange))/numpy.log(base) / dx) + 1 if log else \
                int(numpy.diff(_outRange)/dx) + 1
    _outRange[1] = numpy.power(base, numpy.log(_outRange[0])/numpy.log(base) + dx*(npix-1)) \
                            if log else _outRange[0] + dx*(npix-1)
    return npix, _outRange


class Resample:
    r"""
    Resample regularly or irregularly sampled data to a new grid using
    integration.
    
    This is a generalization of the routine
    :func:`ppxf.ppxf_util.log_rebin` provided by Michele Cappellari in
    the pPXF package.

    The abscissa coordinates (`x`) or the pixel borders (`xBorders`) for
    the data (`y`) should be provided for irregularly sampled data.  If
    the input data is linearly or geometrically sampled (`inLog=True`),
    the abscissa coordinates can be generated using the input range for
    the (geometric) center of each grid point.  If `x`, `xBorders`, and
    `xRange` are all None, the function assumes grid coordinates of
    `x=numpy.arange(y.shape[-1])`.

    The function resamples the data by constructing the borders of the
    output grid using the `new*` keywords and integrating the input
    function between those borders.  The output data will be set to
    `ext_value` for any data beyond the abscissa limits of the input
    data.

    The data to resample (`y`) can be a 1D or 2D vector; the abscissa
    coordinates must always be 1D.  If (`y`) is 2D, the resampling is
    performed along the last axis (i.e., `axis=-1`).

    The nominal assumption is that the provided function is a step
    function based on the provided input (i.e., `step=True`).  If the
    output grid is substantially finer than the input grid, the
    assumption of a step function will be very apparent.  To assume the
    function is instead linearly interpolated between each provided
    point, choose `step=False`; higher-order interpolations are not
    provided.

    If errors are provided, a nominal error propagation is performed to
    provide the errors in the resampled data.  

    .. warning::
        Depending on the details of the resampling, the output errors
        are likely highly correlated.  Any later analysis of the
        resampled function should account for this.  A covariance
        calculation will be provided in the future on a best-effort
        basis.

    The `conserve` keyword sets how the units of the input data should
    be treated.  If `conserve=False`, the input data are expected to be
    in density units (i.e., per `x` coordinate unit) such that the
    integral over :math:`dx` is independent of the units of :math:`x`
    (i.e., flux per unit angstrom, or flux density).  If
    `conserve=True`, the value of the data is assumed to have been
    integrated over the size of each pixel (i.e., units of flux).  If
    `conserve=True`, :math:`y` is converted to units of per step in
    :math:`x` such that the integral before and after the resample is
    the same.  For example, if :math:`y` is a spectrum in units of flux,
    the function first converts the units to flux density and then
    computes the integral over each new pixel to produce the new spectra
    with units of flux.

    .. todo::
        - Allow the user to provide the output pixel borders directly.
        - Allow for higher order interpolations.
        - Allow for a covariance matrix calculation.

    Args:
        y (numpy.ndarray):
            Data values to resample.  Can be a numpy.ma.MaskedArray, and
            the shape can be 1 or 2D.  If 1D, the shape must be
            :math:`(N_{\rm pix},)`; otherwise, it must be
            :math:`(N_y,N_{\rm pix})`.  I.e., the length of the last
            axis must match the input coordinates.
        e (numpy.ndarray, optional):
            Errors in the data that should be resampled.  Can be a
            numpy.ma.MaskedArray, and the shape must match the input `y`
            array.  These data are used to perform a nominal calculation
            of the error in the resampled array.
        mask (numpy.ndarray, optional):
            A boolean array (masked values are True) indicating values
            in `y` that should be ignored during the resampling.  The
            mask used during the resampling is the union of this object
            and the masks of `y` and `e`, if they are provided as
            numpy.ma.MaskedArrays.
        x (numpy.ndarray, optional):
            Abcissa coordinates for the data, which do not need to be
            regularly sampled.  If the pixel borders are not provided,
            they are assumed to be half-way between adjacent pixels, and
            the first and last borders are assumed to be equidistant
            about the provided value.  If these coordinates are not
            provided, they are determined by the input borders, the
            input range, or just assumed to be the indices,
            :math:`0..N_{\rm pix}-1`.
        xRange (array-like, optional):
            A two-element array with the starting and ending value for
            the coordinates of the centers of the first and last pixels
            in y.  Default is :math:`[0,N_{\rm pix}-1]`.
        xBorders (numpy.ndarray, optional):
            An array with the borders of each pixel that must have a
            length of :math:`N_{\rm pix}+1`.
        inLog (:obj:`bool`, optional):
            Flag that the input is logarithmically binned, primarily
            meaning that the coordinates are at the geometric center of
            each pixel and the centers are spaced logarithmically.  If
            false, the sampling is expected to be linear.
        newRange (array-like, optional):
            A two-element array with the (geometric) centers of the
            first and last pixel in the output vector.  If not provided,
            assumed to be the same as the input range.
        newpix (:obj:`int`, optional): 
            Number of pixels for the output vector.  If not provided,
            assumed to be the same as the input vector.
        newLog (:obj:`bool`, optional):
            The output vector should be logarithmically binned.
        newdx (:obj:`float`, optional):
            The sampling step for the output vector.  If `newLog=True`,
            this has to be the change in the logarithm of x for the
            output vector!  If not provided, the sampling is set by the
            output range (see `newRange` above) and number of pixels
            (see `newpix` above).
        base (:obj:`float`, optional):
            The base of the logarithm used for both input and output
            sampling, if specified.  The default is 10; use
            `numpy.exp(1)` for natural logarithm.
        ext_value (:obj:`float`, optional):
            Set extrapolated values to the provided float.  By default,
            extrapolated values are set to 0.  If set to None, values
            are just set to the linear exatrapolation of the data beyond
            the provided limits; use `ext_value=None` with caution!
        conserve (:obj:`bool`, optional):
            Conserve the integral of the input vector.  For example, if
            the input vector is a spectrum in flux units, you should
            conserve the flux in the resampling; if the spectrum is in
            units of flux density, you do not want to conserve the
            integral.
        step (:obj:`bool`, optional):
            Treat the input function as a step function during the
            resampling integration.  If False, use a linear
            interpolation between pixel samples.
    
    Attributes:
        x (numpy.ndarray):
            The coordinates of the function on input.
        xborders (numpy.ndarray):
            The borders of the input pixel samples.
        y (numpy.ndarray):
            The function to resample.
        e (numpy.ndarray):
            The 1-sigma errors in the function to resample.
        m (numpy.ndarray):
            The boolean mask for the input function.
        outx (numpy.ndarray):
            The coordinates of the function on output.
        outborders (numpy.ndarray):
            The borders of the output pixel samples.
        outy (numpy.ndarray):
            The resampled function.
        oute (numpy.ndarray):
            The resampled 1-sigma errors.
        outf (numpy.ndarray):
            The fraction of each output pixel that includes valid data
            from the input function.

    Raises:
        ValueError: Raised if *y* is not of type numpy.ndarray, if *y*
            is not one-dimensional, or if *xRange* is not provided and
            the input vector is logarithmically binned (see *inLog*
            above).
    """
    def __init__(self, y, e=None, mask=None, x=None, xRange=None, xBorders=None, inLog=False,
                 newRange=None, newpix=None, newLog=True, newdx=None, base=10.0, ext_value=0.0,
                 conserve=False, step=True):

        # Check operation can be performed
        if not isinstance(y, numpy.ndarray):
            raise ValueError('Input vector must be a numpy.ndarray!')
        if len(y.shape) > 2:
            raise ValueError('Input must be a 1D or 2D array!')

        # Setup the data, errors, and mask
        self.y = y.filled(0.0) if isinstance(y, numpy.ma.MaskedArray) else y.copy()
        self.twod = self.y.ndim == 2
        self.e = None if e is None \
                    else e.filled(0.0) if isinstance(e, numpy.ma.MaskedArray) else e.copy()
        self.m = numpy.zeros(self.y.shape, dtype=bool) if mask is None else mask

        # Check the shapes
        if self.e is not None and self.e.shape != self.y.shape:
            raise ValueError('Error array shape mismatched!')
        if self.m.shape != self.y.shape:
            raise ValueError('Mask array shape mismatched!')

        # Get the union of all the relevant masks
        if isinstance(y, numpy.ma.MaskedArray):
            self.m |= y.mask
        if e is not None and isinstance(e, numpy.ma.MaskedArray):
            self.m |= e.mask

        # Get the input coordinates
        self.x = None
        self.xborders = None
        # this defines the self.x and self.xborders
        self._input_coordinates(x, xRange, xBorders, inLog, base)

        # If conserving integral, assume input is integrated over pixel
        # width and convert to a density function (divide by pixel size)
        if conserve:
            self.y /= (numpy.diff(self.xborders)[None,:] if self.twod \
                                else numpy.diff(self.xborders))

        # Get the output coordinates
        self.outx = None
        self.outborders = None
        # this defines the self.outx and self.outborders
        self._output_coordinates(newRange, newpix, newLog, newdx, base)

        # Perform the resampling
        self.outy = self._resample_step(self.y) if step else self._resample_linear(self.y)
       
        
        # The mask and errors are always interpolated as a step function
        self.oute = None if self.e is None else self._resample_step(self.e, quad=True)
        
        self.outf = self._resample_step(numpy.invert(self.m).astype(int)) \
                        / numpy.diff(self.outborders)

        # Do not conserve the integral over the size of the pixel
        if not conserve:
            self.outy /= (numpy.diff(self.outborders)[None,:] if self.twod \
                            else numpy.diff(self.outborders))
            if self.oute is not None:
                self.oute /= (numpy.diff(self.outborders)[None,:] if self.twod \
                                    else numpy.diff(self.outborders))

        # Set values for extrapolated regions
        if ext_value is not None:
            indx = (self.outborders[:-1] < self.xborders[0]) \
                        | (self.outborders[1:] > self.xborders[-1]) 
            if numpy.sum(indx) > 0:
                if self.twod:
                    self.outy[:,indx] = ext_value
                    self.outf[:,indx] = 0.
                    if self.oute is not None:
                        self.oute[:,indx] = 0.
                else:
                    self.outy[indx] = ext_value
                    self.outf[indx] = 0.
                    if self.oute is not None:
                        self.oute[indx] = 0.


    def _input_coordinates(self, x, xRange, xBorders, inLog, base):
        """
        Determine the centers and pixel borders of the input
        coordinates.
        """
        if (x is not None or xBorders is not None) and xRange is not None:
            warnings.warn('Provided both x or x borders and the x range.  Ignoring range.')
        _xRange = xRange if x is None and xBorders is None else None
        
        if x is not None:
            if x.ndim != 1:
                raise ValueError('Coordinate vector must be 1D.')
            if x.size != self.y.shape[-1]:
                raise ValueError('Coordinate vector must match last dimension of value array.')
        if xBorders is not None:
            if xBorders.ndim != 1:
                raise ValueError('Coordinate borders must be 1D.')
            if xBorders.size != self.y.shape[-1]+1:
                raise ValueError('Coordinate borders must match last dimension of value array.')

        if x is None:
            if xBorders is not None:
                self.x = numpy.sqrt(xBorders[:-1]*xBorders[1:]) if inLog \
                            else (xBorders[:-1]+xBorders[1:])/2.0
            elif xRange is not None:
                self.x = _pixel_centers(xRange, self.y.shape[-1], log=inLog, base=base)[0]
            else:
                self.x = numpy.arange(self.y.shape[-1]) + 0.5
        else:
            self.x = x

        if xBorders is None:
            dx = numpy.diff(numpy.log(self.x)) if inLog else numpy.diff(self.x)
            self.xborders = numpy.exp(numpy.append(numpy.log(self.x[:-1]) - dx/2,
                                        numpy.log(self.x[-1]) + numpy.array([-1,1])*dx[-1]/2)) \
                                if inLog \
                                else numpy.append(self.x[:-1] - dx/2,
                                                  self.x[-1] + numpy.array([-1,1])*dx[-1]/2)
        else:
            self.xborders = xBorders

    def _output_coordinates(self, newRange, newpix, newLog, newdx, base):
        """Set the output coordinates."""

        # Set the output range and number of pixels
        outRange = numpy.array([self.x[0], self.x[-1]]) if newRange is None \
                        else numpy.array(newRange)
        m, _outRange = resample_vector_npix(outRange=outRange, log=newLog, base=base, dx=newdx,
                                        default=(self.y.shape[-1] if newpix is None else newpix))
        outRange = outRange if _outRange is None else _outRange

        # Get the output pixel borders
        self.outborders = _pixel_borders(outRange, m, log=newLog, base=base)[0]

        # Get the output coordinate vector
        self.outx = numpy.sqrt(self.outborders[:-1]*self.outborders[1:]) if newLog \
                        else (self.outborders[:-1]+self.outborders[1:])/2.0

    def _resample_linear(self, v, quad=False):
        """Resample the vectors."""

        # Combine the input coordinates and the output borders
        combinedX = numpy.append(self.outborders, self.x)
        srt = numpy.argsort(combinedX)
        combinedX = combinedX[srt]

        # Get the indices where the data should be reduced
        border = numpy.ones(combinedX.size, dtype=bool)
        border[self.outborders.size:] = False
        k = numpy.arange(combinedX.size)[border[srt]]

        # Calculate the integrand
        if self.twod:
            # Linearly interpolate the input function at the output border positions
            interp = interpolate.interp1d(self.x, v, axis=-1, assume_sorted=True,
                                          fill_value='extrapolate')
            combinedY = numpy.append(interp(self.outborders), v, axis=-1)[:,srt]
            integrand = (combinedY[:,1:]+combinedY[:,:-1])*numpy.diff(combinedX)[None,:]/2.0
        else:
            # Linearly interpolate the input function at the output border positions
            interp = interpolate.interp1d(self.x, v, assume_sorted=True,
                                          fill_value='extrapolate')
            combinedY = numpy.append(interp(self.outborders), v)[srt]
            integrand = (combinedY[1:]+combinedY[:-1])*numpy.diff(combinedX)/2.0

        if quad:
            integrand = numpy.square(integrand)

        # Use reduceat to calculate the integral
        out = numpy.add.reduceat(integrand, k[:-1], axis=-1) if k[-1] == combinedX.size-1 \
                        else numpy.add.reduceat(integrand, k, axis=-1)[...,:-1]
    
        return numpy.sqrt(out) if quad else out

    def _resample_step(self, v, quad=False):
        """Resample the vectors."""

        # Convert y to a step function
        #  - repeat each element of the input vector twice
        _v = numpy.repeat(v, 2, axis=1) if self.twod else numpy.repeat(v, 2)
        #  - repeat each element of the border array twice, and remove
        #  the first and last elements
        _x = numpy.repeat(self.xborders, 2)[1:-1]

        # Combine the input coordinates and the output borders into a
        # single vector
        indx = numpy.searchsorted(_x, self.outborders)
        combinedX = numpy.insert(_x, indx, self.outborders)

        # Insert points at the borders of the output function
        v_indx = indx.copy()
        v_indx[indx >= _v.shape[-1]] = -1
        combinedY = numpy.array([ numpy.insert(__v, indx, __v[v_indx]) for __v in _v ]) \
                            if self.twod else numpy.insert(_v, indx, _v[v_indx])

        # Calculate the integrand
        integrand = combinedY[:,1:]*numpy.diff(combinedX)[None,:] if self.twod else \
                        combinedY[1:]*numpy.diff(combinedX)
        if quad:
            integrand = numpy.square(integrand)

        # Get the indices where the data should be reduced
        border = numpy.insert(numpy.zeros(_x.size, dtype=bool), indx,
                              numpy.ones(self.outborders.size, dtype=bool))
        k = numpy.arange(combinedX.size)[border]

        # Use reduceat to calculate the integral
        out = numpy.add.reduceat(integrand, k[:-1], axis=-1) if k[-1] == combinedX.size-1 \
                    else numpy.add.reduceat(integrand, k, axis=-1)[...,:-1]
#        if self.twod:
#            out = numpy.array(numpy.add.reduceat(integrand, k[:-1], axis=-1)) \
#                        if k[-1] == combinedX.size-1 else \
#                        numpy.array(numpy.add.reduceat(integrand, k, axis=-1))[:, 0:-1]
#            
#        else:
#            out = numpy.add.reduceat(integrand, k[:-1], axis=-1) if k[-1] == combinedX.size-1 \
#                            else numpy.add.reduceat(integrand, k, axis=-1)[:-1]
            
        
        return numpy.sqrt(out) if quad else out


def rectify_image(img, col, bpm=None, ocol=None, max_ocol=None, extract_width=None,
                  mask_threshold=0.5):
    r"""
    Rectify the image by shuffling flux along columns using the provided
    column mapping.

    The image recification is one dimensional, treating each image row
    independently. It can be done either by a direct resampling of the
    image columns using the provided mapping of output to input column
    location (see `col` and :class:`Resample`) or by an extraction along
    the provided column locations (see `extract_width`). The latter is
    generally faster; however, when resampling each row, the flux is
    explicitly conserved (see the `conserve` argument of
    :class:`Resample`).

    Args:
        img (`numpy.ndarray`_):
            The 2D image to rectify. Shape is :math:`(N_{\rm row},
            N_{\rm col})`.
        col (`numpy.ndarray`_):
            The array mapping each output column to its location in
            the input image. That is, e.g., `col[:,0]` provides the
            column coordinate in `img` that should be rectified to
            column 0 in the output image. Shape is :math:`(N_{\rm
            row}, N_{\rm map})`.
        bpm (`numpy.ndarray`_, optional):
            Boolean bad-pixel mask for pixels to ignore in input
            image. If None, no pixels are masked in the
            rectification. If provided, shape must match `img`.
        ocol (`numpy.ndarray`_, optional):
            The column in the output image for each column in `col`.
            If None, assume::

                ocol = numpy.arange(col.shape[1])

            These coordinates can fall off the output image (i.e.,
            :math:`<0` or :math:`\geq N_{\rm out,col}`), but those
            columns are removed from the output).
        max_ocol (:obj:`int`, optional):
            The last viable column *index* to include in the output
            image; ie., for an image with `ncol` columns, this should
            be `ncol-1`. If None, assume `max(ocol)`.
        extract_width (:obj:`float`, optional):
            The width of the extraction aperture to use for the image
            rectification. If None, the image recification is performed
            using :class:`Resample` along each row.
        mask_threshold (:obj:`float`, optional):
            Either due to `bpm` or the bounds of the provided `img`,
            pixels in the rectified image may not be fully covered by
            valid pixels in `img`. Pixels in the output image with
            less than this fractional coverage of an input pixel are
            flagged in the output.

    Returns:
        Two `numpy.ndarray`_ objects are returned both with shape
        `(nrow,max_ocol+1)`, the rectified image and its boolean
        bad-pixel mask.
    """
    # Check the input
    if img.ndim != 2:
        raise ValueError('Input image must be 2D.')
    if bpm is not None and bpm.shape != img.shape:
        raise ValueError('Image bad-pixel mask must match image shape.')
    _img = numpy.ma.MaskedArray(img, mask=bpm)
    nrow, ncol = _img.shape
    if col.ndim != 2:
        raise ValueError('Column mapping array must be 2D.')
    if col.shape[0] != nrow:
        raise ValueError('Number of rows in column mapping array must match image to rectify.')
    _ocol = numpy.arange(col.shape[1]) if ocol is None else numpy.atleast_1d(ocol)
    if _ocol.ndim != 1:
        raise ValueError('Output column indices must be provided as a vector.')
    if _ocol.size != col.shape[1]:
        raise ValueError('Output column indices must match columns in column mapping array.')
    _max_ocol = numpy.amax(_ocol) if max_ocol is None else max_ocol

    # Use an aperture extraction to rectify the image
    if extract_width is not None:
        # Select viable columns
        indx = (_ocol >= 0) & (_ocol <= _max_ocol)
        # Initialize the output image as all masked
        out_img = numpy.ma.masked_all((nrow,_max_ocol+1), dtype=float)
        # Perform the extraction
        out_img[:,_ocol[indx]] = moment.moment1d(_img.data, col[:,indx], extract_width,
                                                 bpm=_img.mask)[0]
        # Determine what fraction of the extraction fell off the image
        coo = col[:,indx,None] + numpy.arange(numpy.ceil(extract_width)).astype(int)[None,None,:] \
                    - extract_width/2
        in_image = (coo >= 0) | (coo < ncol)
        out_bpm = numpy.sum(in_image, axis=2)/extract_width < mask_threshold
        # Return the filled numpy.ndarray and boolean mask
        return out_img.filled(0.0)/extract_width, out_img.mask | out_bpm

    # Directly resample the image
    # 
    # `col` provides the column in the input image that should
    # resampled to a given position in the output image: the value of
    # the flux at img[:,col[:,0]] should be rectified to `outimg[:,0]`.
    # To run the resampling algorithm we need to invert this. That is,
    # instead of having a function that provides the output column as a
    # function of the input column, we want a function that provies the
    # input column as a function of the output column.

    # Instantiate the output image
    out_img = numpy.zeros((nrow,_max_ocol+1), dtype=float)
    out_bpm = numpy.zeros((nrow,_max_ocol+1), dtype=bool)
    icol = numpy.arange(ncol)
    for i in range(nrow):
        # Get the coordinate vector of the output column
        _icol = interpolate.interp1d(col[i,:], _ocol, copy=False, bounds_error=False,
                                     fill_value='extrapolate', assume_sorted=True)(icol)
        # Resample it
        r = Resample(_img[i,:], x=_icol, newRange=[0,_max_ocol], newpix=_max_ocol+1, newLog=False,
                     conserve=True)
        # Save the resampled data
        out_img[i,:] = r.outy
        # Flag pixels
        out_bpm[i,:] = r.outf < mask_threshold
    return out_img, out_bpm





""" Module for Keck/NIRSPEC specific codes
"""
import numpy as np

from pypeit import msgs
from pypeit.par import pypeitpar
from pypeit.spectrographs import spectrograph
from pypeit import telescopes
from pypeit.core import framematch
from pypeit.par import pypeitpar
from pypeit.spectrographs import spectrograph
from pkg_resources import resource_filename

from pypeit import debugger

class KeckNIRSPECSpectrograph(spectrograph.Spectrograph):
    """
    Child to handle Keck/NIRSPEC specific code
    """
    def __init__(self):
        # Get it started
        super(KeckNIRSPECSpectrograph, self).__init__()
        self.telescope = telescopes.KeckTelescopePar()
        self.camera = 'NIRSPEC'
        self.detector = [
                # Detector 1
            pypeitpar.DetectorPar(
                            dataext         = 0,
                            specaxis        = 0,
                            specflip        = False,
                            xgap            = 0.,
                            ygap            = 0.,
                            ysize           = 1.,
                            platescale      = 0.193,
                            darkcurr        = 0.8,
                            saturation      = 100000.,
                            nonlinear       = 1.00,  # docs say linear to 90,000 but our flats are usually higher
                            numamplifiers   = 1,
                            gain            = 5.8,
                            ronoise         = 23,
                            datasec         = '[:,:]',
                            oscansec        = '[:,:]'
                            )]
        self.numhead = 1
        # Uses default timeunit
        # Uses default primary_hdrext
        # self.sky_file = ?


    def default_pypeit_par(self):
        """
        Set default parameters for Keck/NIRSPEC
        """
        par = pypeitpar.PypeItPar()
        par['rdx']['spectrograph'] = 'keck_nirspec_low'
        # Wavelengths
        # 1D wavelength solution
        par['calibrations']['wavelengths']['rms_threshold'] = 0.20 #0.20  # Might be grating dependent..
        par['calibrations']['wavelengths']['sigdetect']=5.0
        par['calibrations']['wavelengths']['fwhm']= 5.0
        par['calibrations']['wavelengths']['n_final']= 4
        par['calibrations']['wavelengths']['lamps'] = ['OH_NIRES']
        par['calibrations']['wavelengths']['nonlinear_counts'] = self.detector[0]['nonlinear'] * self.detector[0]['saturation']
        par['calibrations']['wavelengths']['method'] = 'holy-grail'
        # Reidentification parameters
        #par['calibrations']['wavelengths']['reid_arxiv'] = 'keck_nires.fits'
        par['calibrations']['slitedges']['edge_thresh'] = 200.
        par['calibrations']['slitedges']['sync_predict'] = 'nearest'

        # Flats
        par['calibrations']['flatfield']['tweak_slits_thresh'] = 0.80
        par['calibrations']['flatfield']['illumflatten'] = True

        # Extraction
        par['reduce']['skysub']['bspline_spacing'] = 0.8
        par['reduce']['extraction']['sn_gauss'] = 4.0

        # Flexure
        par['flexure']['method'] = 'skip'

        par['scienceframe']['process']['sigclip'] = 20.0
        par['scienceframe']['process']['satpix'] ='nothing'


        # Overscan but not bias
        #  This seems like a kludge of sorts
        par['calibrations']['biasframe']['useframe'] = 'none'
        # No overscan
        par['scienceframe']['process']['overscan'] ='none'
        for key in par['calibrations'].keys():
            if 'frame' in key:
                par['calibrations'][key]['process']['overscan'] = 'none'


        # The settings below enable NIRSPEC dark subtraction from the traceframe and pixelflatframe, but enforce
        # that this bias won't be subtracted from other images. It is a hack for now, because eventually we want to
        # perform this operation with the dark frame class, and we want to attach individual sets of darks to specific
        # images.
        par['calibrations']['biasframe']['useframe'] = 'bias'
        par['calibrations']['traceframe']['process']['bias'] = 'force'
        par['calibrations']['pixelflatframe']['process']['bias'] = 'force'
        par['calibrations']['arcframe']['process']['bias'] = 'skip'
        par['calibrations']['tiltframe']['process']['bias'] = 'skip'
        par['calibrations']['standardframe']['process']['bias'] = 'skip'
        par['scienceframe']['process']['bias'] = 'skip'


        # Set the default exposure time ranges for the frame typing
        par['calibrations']['standardframe']['exprng'] = [None, 20]
        par['calibrations']['arcframe']['exprng'] = [20, None]
        par['calibrations']['darkframe']['exprng'] = [20, None]
        par['scienceframe']['exprng'] = [20, None]

        # Sensitivity function parameters
        par['sensfunc']['algorithm'] = 'IR'
        par['sensfunc']['polyorder'] = 8
        par['sensfunc']['IR']['telgridfile'] = resource_filename('pypeit', '/data/telluric/TelFit_MaunaKea_3100_26100_R20000.fits')




        return par

    # JFH Replaced with updated values based on experienced with MOSFIRE above.
    #
    # @staticmethod
    # def default_pypeit_par():
    #     """
    #     Set default parameters for NIRSPEC reductions
    #     """
    #     par = pypeitpar.PypeItPar()
    #     # Frame numbers
    #     par['calibrations']['standardframe']['number'] = 1
    #     par['calibrations']['biasframe']['number'] = 0
    #     par['calibrations']['pixelflatframe']['number'] = 5
    #     par['calibrations']['traceframe']['number'] = 5
    #     par['calibrations']['arcframe']['number'] = 1
    #     # Do not flux calibrate
    #     # NIRSPEC uses sky lines to wavelength calibrate; no need for flexure correction
    #     par['flexure'] = pypeitpar.FlexurePar()
    #     par['flexure']['method'] = 'skip'
    #     # Set the default exposure time ranges for the frame typing
    #     par['calibrations']['arcframe']['exprng'] = [1, None]
    #     par['calibrations']['biasframe']['exprng'] = [None, 2]
    #     par['calibrations']['darkframe']['exprng'] = [None, 5]
    #     par['calibrations']['pinholeframe']['exprng'] = [999999, None]  # No pinhole frames
    #     par['calibrations']['pixelflatframe']['exprng'] = [0, None]
    #     par['calibrations']['traceframe']['exprng'] = [0, None]
    #     par['calibrations']['standardframe']['exprng'] = [None,5]
    #     par['scienceframe']['exprng'] = [1, None]
    #     # Lower the default threshold for tilts
    #     par['calibrations']['tilts']['tracethresh'] = 10.
    #     # Slits
    #     par['calibrations']['slitedges']['edge_thresh'] = 200.
    #     # 1D wavelength solution
    #     par['calibrations']['wavelengths']['lamps']  = ['OH_R24000']
    #     par['calibrations']['wavelengths']['rms_threshold'] = 0.20  # Good for NIRSPEC-1
    #     par['calibrations']['wavelengths']['sigdetect'] = 5.      # Good for NIRSPEC-1
    #
    #     return par
    #

    '''
    def check_headers(self, headers):
        """
        Check headers match expectations for an LRISb exposure.
        See also
        :func:`pypeit.spectrographs.spectrograph.Spectrograph.check_headers`.
        Args:
            headers (list):
                A list of headers read from a fits file
        """
        expected_values = { '0.INSTRUME': 'NIRSPEC',
                               '0.NAXIS': 2,
                              '0.NAXIS1': 1024,
                              '0.NAXIS2': 1024 }
        super(KeckNIRSPECSpectrograph, self).check_headers(headers,
                                                           expected_values=expected_values)

    def header_keys(self):
        """
        Return a dictionary with the header keywords to read from the
        fits file.
        Returns:
            dict: A nested dictionary with the header keywords to read.
            The first level gives the extension to read and the second
            level gives the common name for header values that is passed
            on to the PypeItMetaData object.
        """
        hdr_keys = {}
        hdr_keys[0] = {}
        hdr_keys[0]['idname'] = 'IMAGETYP'
        #hdr_keys[0]['date'] = 'DATE-OBS'
        hdr_keys[0]['utc'] = 'UTC'
        hdr_keys[0]['target'] = 'OBJECT'
        hdr_keys[0]['time'] = 'MJD-OBS'
        hdr_keys[0]['ra'] = 'RA'
        hdr_keys[0]['dec'] = 'DEC'
        hdr_keys[0]['airmass'] = 'AIRMASS'
        hdr_keys[0]['decker'] = 'SLITNAME'
        hdr_keys[0]['echellepos'] = 'ECHLPOS'
        hdr_keys[0]['crosspos'] = 'DISPPOS'
        hdr_keys[0]['naxis0'] = 'NAXIS2'
        hdr_keys[0]['naxis1'] = 'NAXIS1'
        hdr_keys[0]['filter1'] = 'FILNAME'
        hdr_keys[0]['dispname'] = 'DISPERS'
        hdr_keys[0]['hatch'] = 'CALMPOS'
        hdr_keys[0]['slitwid'] = 'SLITWIDT'
        hdr_keys[0]['slitlen'] = 'SLITLEN'

        # 'ELAPTIME' is added by KOA, but otherwise would need to do 'ITIME' * 'COADDS'
        hdr_keys[0]['exptime'] = 'ELAPTIME'

        # Lamp names and statuses
        lamp_names = ['NEON', 'ARGON', 'KRYPTON', 'XENON', 'ETALON', 'FLAT']
        for kk,lamp_name in enumerate(lamp_names):
            hdr_keys[0]['lampstat{:02d}'.format(kk+1)] = lamp_name

        return hdr_keys
    '''
    def init_meta(self):
        """
        Generate the meta data dict
        Note that the children can add to this

        Returns:
            self.meta: dict (generated in place)

        """
        meta = {}
        # Required (core)
        meta['ra'] = dict(ext=0, card='RA')
        meta['dec'] = dict(ext=0, card='DEC')
        meta['target'] = dict(ext=0, card='TARGNAME')
        meta['decker'] = dict(ext=0, card='SLITNAME')
        meta['binning'] = dict(ext=0, card=None, default='1,1')

        meta['mjd'] = dict(ext=0, card='MJD-OBS')
        meta['exptime'] = dict(ext=0, card='ELAPTIME')
        meta['airmass'] = dict(ext=0, card='AIRMASS')
        # Extras for config and frametyping
        meta['dispname'] = dict(ext=0, card='DISPERS')
        meta['hatch'] = dict(ext=0, card='CALMPOS')
        meta['idname'] = dict(ext=0, card='IMAGETYP')
        # Lamps
        lamp_names = ['NEON', 'ARGON', 'KRYPTON', 'XENON', 'ETALON', 'FLAT']
        for kk,lamp_name in enumerate(lamp_names):
            meta['lampstat{:02d}'.format(kk+1)] = dict(ext=0, card=lamp_name)
        # Ingest
        self.meta = meta

    def configuration_keys(self):
        return ['decker', 'dispname']

    '''
    def metadata_keys(self):
        return super(KeckNIRSPECSpectrograph, self).metadata_keys() \
                    + ['echellepos', 'crosspos', 'idname']
    '''
    def pypeit_file_keys(self):
        pypeit_keys = super(KeckNIRSPECSpectrograph, self).pypeit_file_keys()
        pypeit_keys += ['calib', 'comb_id', 'bkg_id']
        return pypeit_keys

    def check_frame_type(self, ftype, fitstbl, exprng=None):
        """
        Check for frames of the provided type.
        """
        good_exp = framematch.check_frame_exptime(fitstbl['exptime'], exprng)
        if ftype in ['science', 'standard']:
            return good_exp & self.lamps(fitstbl, 'off') & (fitstbl['hatch'] == 0) \
                        & (fitstbl['idname'] == 'object')
        if ftype in ['bias', 'dark']:
            return good_exp & self.lamps(fitstbl, 'off') & (fitstbl['hatch'] == 0) \
                        & (fitstbl['idname'] == 'dark')
        if ftype in ['pixelflat', 'trace']:
            # Flats and trace frames are typed together
            return good_exp & self.lamps(fitstbl, 'dome') & (fitstbl['hatch'] == 1) \
                        & (fitstbl['idname'] == 'flatlamp')
        if ftype == 'pinhole':
            # Don't type pinhole frames
            return np.zeros(len(fitstbl), dtype=bool)
        if ftype in ['arc', 'tilt']:
            # TODO: This is a kludge.  Allow science frames to also be
            # classified as arcs
            is_arc = self.lamps(fitstbl, 'arcs') & (fitstbl['hatch'] == 1) \
                            & (fitstbl['idname'] == 'arclamp')
            is_obj = self.lamps(fitstbl, 'off') & (fitstbl['hatch'] == 0) \
                        & (fitstbl['idname'] == 'object')
            return good_exp & (is_arc | is_obj)
        msgs.warn('Cannot determine if frames are of type {0}.'.format(ftype))
        return np.zeros(len(fitstbl), dtype=bool)

    def lamps(self, fitstbl, status):
        """
        Check the lamp status.
        Args:
            fitstbl (:obj:`astropy.table.Table`):
                The table with the fits header meta data.
            status (:obj:`str`):
                The status to check.  Can be `off`, `arcs`, or `dome`.
        Returns:
            numpy.ndarray: A boolean array selecting fits files that
            meet the selected lamp status.
        Raises:
            ValueError:
                Raised if the status is not one of the valid options.
        """
        if status == 'off':
            # Check if all are off
            return np.all(np.array([fitstbl[k] == 0 for k in fitstbl.keys() if 'lampstat' in k]),
                          axis=0)
        if status == 'arcs':
            # Check if any arc lamps are on
            arc_lamp_stat = [ 'lampstat{0:02d}'.format(i) for i in range(1,6) ]
            return np.any(np.array([ fitstbl[k] == 1 for k in fitstbl.keys()
                                            if k in arc_lamp_stat]), axis=0)
        if status == 'dome':
            return fitstbl['lampstat06'] == 1

        raise ValueError('No implementation for status = {0}'.format(status))

    # TODO: This function is unstable to shape...
    def bpm(self, filename, det, shape=None, msbias=None):
        """ Generate a BPM
        Parameters
        ----------
        shape : tuple, REQUIRED
        msbias : numpy.ndarray, required if the user wishes to generate a BPM based on a master bias

        Returns
        -------
        badpix : ndarray
        """
        bpm_img = self.empty_bpm(filename, det)

        # Fill in bad pixels if a master bias frame is provided
        if msbias is not None:
            return self.bpm_frombias(msbias, det, bpm_img)

        # Edges of the detector are junk
        msgs.info("Custom bad pixel mask for NIRSPEC")
        bpm_img[:, :20] = 1.
        bpm_img[:, 1000:] = 1.

        return bpm_img

class KeckNIRSPECLowSpectrograph(KeckNIRSPECSpectrograph):
    """
    Child to handle NIRSPEC low-dispersion specific code
    """

    def __init__(self):
        # Get it started
        super(KeckNIRSPECLowSpectrograph, self).__init__()
        self.spectrograph = 'keck_nirspec_low'


    @property
    def telluric_grid_file(self):
        """Return the grid of HITRAN atmosphere models for telluric correctinos"""
        return resource_filename('pypeit', '/data/telluric/TelFit_MaunaKea_3100_26100_R20000.fits')



    '''
    def check_header(self, headers):
        """Validate elements of the header."""
        chk_dict = {}
        # chk_dict is 1-indexed!
        chk_dict[1] = {}
        # THIS CHECK IS A MUST! It performs a standard check to make sure the data are 2D.
        chk_dict[1]['NAXIS'] = 2
        return chk_dict
    '''





"""
Module to compute moments.

.. include common links, assuming primary doc root is up one directory
.. include:: ../links.rst
"""

import numpy as np
from scipy import special

def moment1d(flux, col, width, ivar=None, bpm=None, fwgt=None, row=None, weighting='uniform',
             order=0, bounds=None, fill_error=-1., mesh=False):
    r"""
    Compute one-dimensional moments of the provided image within an
    aperture along its second axis (axis=1).

    This method allows for computations of the zeroth, first, and
    second moments (see `order`). The aperture used for the
    calculation is centered at the provided `col` pixel with a width
    defined by `width`; however, this definition depends on the type
    of weighting applied (see `weighting`). Formulae for each moment
    are as follows. The zeroth moment (`order=0`) computes the
    discrete weighted sum of the flux within the aperture:
    
    .. math::
        \mu_0 &= \sum_i w_i f_i \\
        \epsilon_{\mu_0}^2 &= \sum_i (w_i \epsilon_{f,i})^2,

    where :math:`f` is the flux in each pixel :math:`i`,
    :math:`\epsilon_f` is its error, and :math:`w` is the assigned
    weight (see `weighting`).  The first moment (`order=1`) computes the
    flux-weighted center of window:

    .. math::
        \mu_1 &= \frac{\sum_i x_i w_i f_i }{\mu_0} \\
        \epsilon_{\mu_1}^2 &= \mu_0^{-2}\ \sum_i [ w_i \epsilon_{f,i}\
                            (x_i - \mu_1)]^2,

    where :math:`x` is the pixel position along the 2nd axis (see
    `col`).  The second moment (`order=2`) computes the variance of the
    flux profile about its center within the window:
    
    .. math::
        \mu_2 &= \frac{\sum_i x^2_i w_i f_i }{\mu_0} - \mu_1^2 \\
        \epsilon_{\mu_2}^2 &= \mu_0^{-2}\ \sum_i w^2_i \epsilon^2_{f,i}\
                                [(x_i - \mu_1)^2 + \mu_2]^2.

    The values returned for the second-moment calculation are actually
    the standard deviation instead of the variance, where:

    .. math::
        \sigma = \sqrt{\mu_2} \\
        \epsilon_{\sigma} = \frac{\epsilon_{\mu_2}}{2 \sigma}.

    The method uses `numpy.ma.MaskedArray` objects to keep track of
    math errors, such as divisions by 0. The returned boolean array
    indicates when these errors occurred, and the method replaces
    these errors with the original centers.

    The shape of arrays depend on the shapes of the input image
    (`flux`), columns about which to perform the calculation (`col`),
    and rows from which to select the column data (`row`) as listed
    below.  In the following, the input flux array must be 2D and has
    shape :math:`(N_{\rm row}, N_{\rm col})`, whereas the number of rows
    and columns identified for a moment calculation are :math:`N_{\rm
    mom,row}` and :math:`N_{\rm mom,col}`, respectively.  The possible
    shapes of the output arrays are:

        - If `row` is None and `col` is 1D, the moments are calculated
          centered at each `col` for all rows in `flux`. The shape of
          the array for each moment is then :math:`(N_{\rm row}, N_{\rm
          mom,col})`
        - If `row` is None and `col` is 2D, `col` must have shape
          :math:`(N_{\rm row}, N_{\rm mom,col})` and the shape of the
          output array per moment calculated is the same.
        - If `row` is a single integer and `col` is 1D, the moments are
          calculated for each col only at the provided row in `flux`.
          The shape of the array for each moment is then :math:`(N_{\rm
          mom,col},)`.
        - If `row` is a 1D integer array and `col` is 1D, the default
          behavior is for the moments to be calculated at each col
          for each provided row. The shape of the array for each
          moment would then be :math:`(N_{\rm mom,row},N_{\rm
          mom,col})`. This will be true even if :math:`N_{\rm
          mom,row} = N_{\rm mom,col}`, as long as `mesh` is `True`.
          However, if :math:`N_{\rm mom,row} = N_{\rm mom,col}` and
          `mesh` is `False`, only the matched pairs of `col` and
          `row` will be used and the output shape per moment will be
          :math:`(N_{\rm mom,row},)`.
        - If `row` is a 1D integer array and `col` is 2D, the first axis
          of `col` must be the same length as `row`. The shape of the
          array for each moment is then :math:`(N_{\rm mom,row},N_{\rm
          mom,col})`, which is the same as the input `col`.
        - If `row` is 2D and `col` is 2D, they must have the same shape.
        - If `row` is a single integer and `col` is 2D, or if `row` is
          2D and `col` is 1D, the method raises an error.

    .. note::

        - This is an entirely general function, as reflected by the
          nomenclature used in the call. As used within PypeIt, the
          PypeIt image orientation convention means that moments are
          always taken along the spatial direction; i.e., `col` is
          the spatial coordinate and `row` is the spectral
          coordinate.

        - This function is a generalization of and builds on the
          heritage of functions in idlspec2d, specifically
          trace_fweight, trace_gweight, extrace_asymbox2,
          extract_boxcar.

    .. warning::

        The function has significant setup/input checking. Most of
        this introduces limited overhead with the exception of the
        handling of `ivar`, `bpm`, and `fwgt`. If any of these are
        provided as `None` on input, an array is constructed (unity
        for `ivar` and `fwgt` and all False for `bpm`) that serves as
        a place holder. If repetitive calls to the function are
        expected and any of these arrays are missing, significant
        efficiency gains can be made by providing pre-built values
        for these arrays so that time isn't lost in allocating the
        placeholder arrays in every call.

    .. todo::

        Optimize the code for efficiency, regardless of the input.

    Args:
        flux (`numpy.ndarray`_):
            Intensity image with shape :math:`(N_{\rm row}, N_{\rm
            col})`.
        col (`numpy.ndarray`_):
            Floating-point center along the 2nd axis for the integration
            window in pixel index (first value located at index 0).
            This can either be a 1D or 2D array.  See restrictions on
            the shape in the description above.
        width (:obj:`float`, `numpy.ndarray`_):
            The meaning of the parameter depends on the value of
            `weighting`.  If `weighting=='uniform'`, the width of the
            integration window in columns, centered at the input `col`.
            If `weighting=='gaussian'`, the :math:`\sigma` of a
            pixelated Gaussian weighting function.  The width of the
            integration window in columns, centered at the input `col`,
            is always `6*width` (i.e., the half-width is
            :math:`3\sigma`).  The provided value can be a scalar to use
            a constant window definition, or it can be an array variable
            integration window where the array must have the same shape
            as `col`.
        ivar (`numpy.ndarray`_, optional):
            Inverse variance of the image intensity.  If not provided,
            unity variance is used.  If provided, must have the same
            shape as `flux`.
        bpm (`numpy.ndarray`_, optional):
            Boolean bad-pixel mask for the input image. True values
            are ignored, False values are included. If not provided,
            all pixels are included. If provided, must have the same
            shape as `flux`.
        fwgt (`numpy.ndarray`_, optional):
            An additional weight to apply to each pixel in `flux`.  If
            None, weights are uniform.  Otherwise, the :math:`w_i` from
            above are the product of this weight and the result of the
            scheme set using the `weighting` argument.
        row (:obj:`int`, `numpy.ndarray`_, optional):
            Integer or integer array with the position along the first
            axis (axis=0) for the moment calculation.  This can either
            be None, an integer, or a 1D or 2D array.  See restrictions
            on the shape in the description above.
        weighting (:obj:`str`, optional):
            The weighting to apply to the position within each
            integration window (see `width` above). This must be
            (case-insensitive) either 'uniform' for uniform weighting or
            'gaussian' for weighting by a Gaussian centered at the input
            guess coordinates and integrated over the pixel width.
        order (:obj:`int`, array-like, optional):
            The order of the moment(s) to calculate. Can be a single
            integer or a list. Moments to calculate must be 0, 1, or
            2; at most order can be `[0,1,2]`. The shape of the
            output arrays depends on the number of moments
            calculated. Note that the calculation of the orders is
            necessarily sequential; i.e., setting `order=2` means
            that the zeroth and first moments have to be calculated
            anyway. The order must be provided in sorted order; i.e.,
            you cannot pass `order=[2,1]`.
        bounds (:obj:`tuple`, optional):
            A two-tuple with the lower and upper limit for each
            moment order. If None, no bounds are imposed. If not
            None, an upper and lower bound must be provided for each
            moment to compute; i.e., if more than one moment is
            computed, each element of the two-tuple must be an
            array-like object that matches the length of `order`. To
            set an upper or lower bound only, set the unbounded
            component to None. Bounds for the zeroth and second order
            moments are in an absolute sense, whereas first-order
            bounds are relative to the input `col`. Measurements that
            hit the bounds are masked; see the description of the
            returned objects. For example, to flag anything without a
            positive zeroth moment or a maximum shift from the input
            center of 1 pixel, call the method with arguments::

                order=[0,1], bounds=([0,-1], [None,1])
            
        fill_error (:obj:`float`, optional):
            Value to use as filler for undetermined moments,
            resulting from either the input bad-pixel mask or
            computational issues (division by zero, etc.; see return
            description below).
        mesh (:obj:`bool`, optional):
            If `col` and `row` are 1D vectors of the same length,
            this determines if each `col` and `row` should be paired
            (`mesh is False`) or used to construct a grid, i.e.,
            every `col` combined with every `row` (`mesh is True`).
            See the method description.

    Returns:
        Three `numpy.ndarray`_ objects are returned. If more than one
        moment order is requested, the moments are ordered along the
        first axis; e.g., if `order=[0,1]` the outputs `moment[0]` and
        `moment[1]` contain the zeroth and first moments, respectively.
        The subsequent dimensions of the output arrays are dictated by
        the input `row` and `col`; see the method description. The
        returned arrays are:

            - The moment calculated along the 2nd axis of the input
              image (axis=1).  Masked values (indicated by the third
              object returned) are 0 for the zeroth and second moments
              and equal to the input `col` value for the first moments.
            - The formal propagated error (see equations above) in the
              moments.  Errors are only meaningful if `ivar` is
              provided. Masked values (indicated by the third object
              returned) are set to `fill_error`.
            - A boolean bad-pixel mask for output data; True values
              should be ignored, False values are valid measurements.

    Raises:
        ValueError:
            Raised if input shapes are not correct or if the selected
            `weighting` is unknown.  See method description.

    Examples:

        First setup an image with some Gaussians:

        >>> from pypeit.core.moment import moment1d
        >>> import numpy as np
        >>> from scipy.special import erf
        >>> def gauss_comb():
        ...     c = [45,50,55]
        ...     img = np.zeros((len(c),100), dtype=float)
        ...     x = np.arange(100)
        ...     sig = 5.
        ...     for i,_c in enumerate(c):
        ...         img[i,:] = (erf((x-c[i]+0.5)/np.sqrt(2)/sig) 
        ...                         - erf((x-c[i]-0.5)/np.sqrt(2)/sig))/2.
        ...     return img
        ...
        >>> img = gauss_comb()

        Calculate all moments at one column and row:
        
        >>> mu, mue, flag = moment1d(img, 50, 40., row=0, order=[0,1,2])
        >>> print(mu)
        [ 0.99858297 45.02314924  4.97367636]
    
        Calculate all moments at one column for all rows:
        
        >>> moment1d(img, 50, 40., order=[0,1,2])[0]
        array([[ 0.99858297,  0.99993125,  0.99858297],
               [45.02314924, 50.        , 54.97685076],
               [ 4.97367636,  5.00545947,  4.97367636]])
    
        Calculate zeroth moments in all rows centered at column 50
        
        >>> moment1d(img, 50, 40., order=0)[0]
        array([0.99858297, 0.99993125, 0.99858297])

        Calculate zeroth moments in all rows for three column positions
        
        >>> moment1d(img, [45,50,55], 40., order=0, mesh=True)[0]
        array([[0.99993125, 0.99858297, 0.97670951],
               [0.99858297, 0.99993125, 0.99858297],
               [0.97670951, 0.99858297, 0.99993125]])

        Calculate first moments in all rows for the three column positions
        
        >>> moment1d(img, [45,50,55], 40., order=1, mesh=True)[0]
        array([[45.        , 45.02314924, 45.2814655 ],
               [49.97685076, 50.        , 50.02314924],
               [54.7185345 , 54.97685076, 55.        ]])

        Or pick the same column for all rows
        
        >>> moment1d(img, 50, 40., row=[0,1,2], order=1)[0]
        array([45.02314924, 50.        , 54.97685076])

        Or pick a column unique to each row
        
        >>> moment1d(img, [43,52,57], 40., row=[0,1,2], order=1)[0]
        array([44.99688181, 50.00311819, 55.00311819])

    """

    # TODO: Could be generalized further for higher dimensional
    # inputs...

    # TODO: Need to benchmark again after including the `bounds`
    # keyword.

    # Check mode input
    _weighting = weighting.lower()
    if _weighting not in ['uniform', 'gaussian']:
        raise ValueError('Weighting must be uniform or gaussian')

    # Check image input
    # TODO: For large images, instantiating ivar, bpm, and fwgt can be
    # a *significant* time sink if there are only a few moments
    # calculated
    if flux.ndim != 2:
        raise ValueError('Input image must be 2D.')
    nrow, ncol = flux.shape
    if ivar is not None and ivar.shape != flux.shape:
        raise ValueError('Inverse variance must have the same shape as the input image.')
    if bpm is not None and bpm.shape != flux.shape:
        raise ValueError('Pixel bad-pixel mask must have the same shape as the input image.')
    if fwgt is not None and fwgt.shape != flux.shape:
        raise ValueError('Pixel weights must have the same shape as the input image.')

    # Check moment order
    _order = np.atleast_1d(order)
    if not np.array_equal(_order, np.sort(_order)):
        # NOTE: This is rather strict, but it's needed to make sure
        # that the provided order and bounds make sense.
        raise ValueError('Order must be provided as a sorted array.')
    if _order.ndim != 1:
        raise ValueError('Order can be at most 1D.')
    if _order.size > 3:
        raise ValueError('Can return at most 3 moments.')
    if not np.array_equal(np.unique(_order), _order):
        raise ValueError('Moments must be unique!')
    if np.any((_order != 0) & (_order != 1) & (_order != 2)):
        raise ValueError('Selected moments must be either 0, 1, or 2.')
    norder = len(_order)

    # Check if the bounds are provided
    lower = np.array([None, None, None], dtype=object)
    upper = np.array([None, None, None], dtype=object)
    if bounds is not None:
        if len(bounds) != 2:
            raise ValueError('Bounds must be provided as a two-tuple.')
        _bounds = tuple([np.atleast_1d(b) for b in bounds])
        if np.any([len(b) != norder for b in _bounds]):
            raise ValueError('Number of bounds must match number of moments to calculate.')
        lower[_order] = _bounds[0]
        upper[_order] = _bounds[1]

    # Check coordinate input.  For both col and width, the atleast_1d
    # function does not result in a copy if the provided object is a
    # numpy.ndarray
    array_input = isinstance(col, np.ndarray)
    _col = np.atleast_1d(col)
    _width = np.atleast_1d(width)
    if _width.size == 1:
        _width = np.full(_col.shape, width, dtype=float)
    if _width.shape != _col.shape:
        raise ValueError('width must either be a single value or have the same shape as col.')
    if _col.ndim > 2:
        raise ValueError('Input columns for calculation must be a 1D or 2D array.')
    if row is None and _col.ndim == 2 and _col.shape[0] != nrow:
        raise ValueError('Length of first axis of col and flux must match.')
    if row is not None:
        _row = np.atleast_1d(row)
        if _row.ndim == 1 and _col.ndim == 2 and _col.shape[0] != _row.size:
            raise ValueError('First axis of col must match length of row.')
        if _row.ndim == 2 and _col.ndim == 2 and _col.shape != _row.shape:
            raise ValueError('col and row must have the same shape.')
        if len(_row) == 1 and _col.ndim == 2 or _row.ndim == 2 and _col.ndim == 1:
            raise ValueError('Invalid combination of row and col input shapes.')
    else:
        _row = np.arange(nrow)
    # Check column and row values are valid

    # TODO: This allows the window centers to be outside the image
    # range. This is okay because _col is never used when slicing the
    # image. However, _row values are, meaning that they have to be
    # within the image limits.
#    if np.any((_col < 0) | (_col >= ncol)):
#        raise ValueError('Column locations outside provided image.')
    if np.any((_row < 0) | (_row >= nrow)):
        raise ValueError('Row locations outside provided image.')

    # Fill the columns and rows so that they have matching shape
    rmrowdim = _row.size == 1
    rmcoldim = _col.size == 1 and not array_input
    if _col.ndim == 1 and _row.ndim == 1:
        if _row.size == 1 or _row.size == _col.size and not mesh:
            if _row.size == 1:
                # Repeat the row for each column
                _row = np.full(_col.size, _row[0], dtype=int)
            else:
                # mesh is False, so just match column and row numbers
                _row.reshape(1,-1)
                rmrowdim = True
            _col = _col.reshape(1,-1)
            _width = _width.reshape(1,-1)
        else:
            # Do the combinatorics for rows, columns, and integration
            # widths for the measurements
            nmomcol = _col.size
            _col = np.tile(_col, (_row.size,1))
            _width = np.tile(_width, (_row.size,1))
            _row = np.tile(_row, (nmomcol,1)).T
    elif _col.ndim == 2 and _row.ndim == 1:
        _row = np.tile(_row, (_col.shape[1],1)).T

    # Construct the shape for the output
    outshape = (norder,) + _col.shape
    if rmrowdim and rmcoldim:
        outshape = (outshape[0],)
    elif rmrowdim:
        outshape = (outshape[0],outshape[2])
    elif rmcoldim:
        outshape = (outshape[0],outshape[1])
    if norder == 1 and len(outshape) > 1:
        outshape = outshape[1:]

    # Flatten the coordinate arrays (creates a copy).
    _row = _row.flatten().astype(int)
    _col = _col.flatten().astype(float)
    _width = _width.flatten().astype(float)

    # The "radius" of the pixels to cover is either half of the
    # provided width for uniform weighting or 3*width for Gaussian
    # weighting, where width is the sigma of the Gaussian
    _radius = _width/2 if _weighting == 'uniform' else _width*3

    # Window for the integration for each coordinate. In the
    # calculation of `c`, the increase of the window size by 4 isn't
    # strictly necessary. At minimum it has to be 2, but the increase
    # by 4 ensures there are 0 pixels at either end of the integration
    # window. TODO: Should consider changing this to 2.
    i1 = np.floor(_col - _radius + 0.5).astype(int)
    i2 = np.floor(_col + _radius + 0.5).astype(int)
    c = i1[:,None]-1+np.arange(int(np.amax(np.amin(i2-i1)-1,0))+4)[None,:]
    ih = np.clip(c,0,ncol-1)

    # Set the weight over the window; masked pixels have 0 weight
    good = (c >= 0) & (c < ncol)
    if bpm is not None:
        # NOTE: `&=` doesn't work here because of the np.newaxis usage
        good = good & np.invert(bpm[_row[:,None],ih])
    if ivar is not None:
        good = good & (ivar[_row[:,None],ih] > 0)
    if _weighting == 'uniform':
        # Weight according to the fraction of each pixel within in the
        # integration window
        wt = good * np.clip(_radius[:,None] - np.abs(c - _col[:,None]) + 0.5,0,1)
    else:
        # Weight according to the integral of a Gaussian over the pixel
        coo = c - _col[:,None]
        wt = good * (special.erf((coo+0.5)/np.sqrt(2.)/_width[:,None])
                        - special.erf((coo-0.5)/np.sqrt(2.)/_width[:,None]))/2.

    # Construct the moment-independent component of the integrand and
    # the zeroth moment; the zeroth moment is always needed
    integ = flux[_row[:,None],ih] * wt
    if fwgt is not None:
        integ *= fwgt[_row[:,None],ih]
    mu = np.array([np.ma.sum(integ, axis=1), None, None], dtype=object)
    mue = np.array([None, None, None], dtype=object)
    mum = np.array([np.ma.getmaskarray(mu[0]).copy(), None, None], dtype=object)
    _var0 = np.square(wt) if ivar is None else np.ma.divide(np.square(wt), ivar[_row[:,None],ih])
    if 0 in _order:
        # Only calculate the error if the moment was requested
        mue[0] = np.ma.sqrt(np.ma.sum(_var0, axis=1))
        # Impose the boundary
        if lower[0] is not None:
            mum[0] |= mu[0] < lower[0]
        if upper[0] is not None:
            mum[0] |= mu[0] > upper[0]
    
    # Calculate the first moment if necessary
    if np.any(_order > 0):
        mu[1] = np.ma.divide(np.sum(integ*c, axis=1), mu[0])
        mum[1] = np.ma.getmaskarray(mu[1]).copy()
        if 1 in _order:
            # Only calculate the error if the moment was requested
            _var1 = np.square(wt * (c - mu[1][:,None]))
            if ivar is not None:
                _var1 = np.ma.divide(_var1, ivar[_row[:,None],ih])
            mue[1] = np.ma.divide(np.ma.sqrt(np.ma.sum(_var1, axis=1)), np.absolute(mu[0]))
            # Impose the boundary
            if lower[1] is not None:
                mum[1] |= mu[1] < _col - lower[1]
            if upper[1] is not None:
                mum[1] |= mu[1] > _col + upper[1]

    # Calculate the second moment if necessary
    if 2 in _order:
        mu[2] = np.ma.divide(np.sum(integ*np.square(c), axis=1), mu[0]) - np.square(mu[1])
        mue[2] = np.ma.divide(np.ma.sqrt(
                        np.ma.sum(_var0 * np.square(np.square(c - mu[1][:,None]) + mu[2][:,None]),
                                  axis=1)), np.absolute(mu[0]))
        mu[2] = np.ma.sqrt(mu[2])
        mue[2] = np.ma.divide(mue[2], 2*mu[2])
        mum[2] = np.ma.getmaskarray(mu[2]).copy()
        # Impose the boundary
        if lower[2] is not None:
            mum[2] |= mu[2] < lower[2]
        if upper[2] is not None:
            mum[2] |= mu[2] > upper[2]

    # Fill in the masked values
    for i in range(3):
        if mu[i] is None:
            continue
        mu[i][mum[i]] = _col[mum[i]] if i == 1 else 0.0
        mu[i] = mu[i].data
        if mue[i] is None:
            continue
        mue[i][mum[i]] = fill_error
        mue[i] = mue[i].filled(fill_error)

    # Return with the correct shape
    singlenum = outshape == (1,) and not array_input
    return (mu[_order][0][0], mue[_order][0][0], mum[_order][0][0]) if singlenum \
            else (np.concatenate(mu[_order]).reshape(outshape),
                  np.concatenate(mue[_order]).reshape(outshape),
                  np.concatenate(mum[_order]).reshape(outshape))




"""
Module to run tests on core.procimg functions.
"""
import pytest
import numpy as np

from pypeit.core import procimg

def test_replace_columns():
    y = np.zeros((10,3), dtype=float)
    y[:,2] = 2
    bad_col = np.array([False, True, False])
    assert np.array_equal(procimg.replace_columns(y, bad_col, copy=True, replace_with='mean'),
                          procimg.replace_columns(y, bad_col, copy=True, replace_with='linear')), \
                'Interpolation and mean should provide the same result.'

    bad_col = np.array([False, True, True])
    assert np.array_equal(procimg.replace_columns(y, bad_col, copy=True, replace_with='mean'),
                          np.zeros_like(y)), 'Should set everything to 0.'

    bad_col = np.array([True, True, False])
    assert np.array_equal(procimg.replace_columns(y, bad_col, copy=True, replace_with='mean'),
                          np.full_like(y, 2)), 'Should set everything to 2.'

    y = np.zeros((10,4), dtype=float)
    y[:,3] = 3
    bad_col = np.array([False, True, True, False])
    assert np.array_equal(procimg.replace_columns(y, bad_col, copy=True, replace_with='linear'),
                          np.repeat(np.arange(4),10).reshape(4,10).T), \
                'Interpolation failed.'




"""
Script for performing 2d coadds of PypeIt data.
"""
import os
import glob
import argparse
from collections import OrderedDict

import numpy as np

from configobj import ConfigObj

from astropy.io import fits

from pypeit.pypeit import PypeIt
from pypeit import par, msgs
from pypeit import coadd2d
from pypeit.core import save
from pypeit import io
from pypeit.spectrographs.util import load_spectrograph


# TODO: We need an 'io' module where we can put functions like this...
def read_coadd2d_file(ifile):
    """
    Read a PypeIt coadd2d file, akin to a standard PypeIt file.

    .. todo::

        - Need a better description of this.  Probably for the PypeIt
          file itself, too!

    The top is a config block that sets ParSet parameters.  The
    spectrograph is required.

    Args:
        ifile (:obj:`str`):
          Name of the flux file

    Returns:
        tuple: Returns three objects: (1) The
        :class:`pypeit.spectrographs.spectrograph.Spectrograph`
        instance, (2) the list of configuration lines used to modify the
        default :class`pypeit.par.pypeitpar.PypeItPar` parameters, and
        (3) the list of spec2d files to combine.
    """

    # Read in the pypeit reduction file
    msgs.info('Loading the coadd2d file')
    lines = par.util._read_pypeit_file_lines(ifile)
    is_config = np.ones(len(lines), dtype=bool)


    # Parse the coadd block
    spec2d_files = []
    s, e = par.util._find_pypeit_block(lines, 'coadd2d')
    if s >= 0 and e < 0:
        msgs.error("Missing 'coadd2d end' in {0}".format(ifile))
    for line in lines[s:e]:
        prs = line.split(' ')
        # TODO: This needs to allow for the science directory to be
        # defined by the user.
        #spec2d_files.append(os.path.join(os.path.basename(prs[0])))
        spec2d_files.append(prs[0])
    is_config[s-1:e+1] = False
    # Construct config to get spectrograph
    cfg_lines = list(lines[is_config])
    cfg = ConfigObj(cfg_lines)
    spectrograph_name = cfg['rdx']['spectrograph']
    spectrograph = load_spectrograph(spectrograph_name)

    # Return
    return spectrograph, cfg_lines, spec2d_files


def parser(options=None):
    parser = argparse.ArgumentParser(description='Parse')
    parser.add_argument("--file", type=str, default=None, help="File to guide 2d coadds")
    parser.add_argument('--det', default=None, type=int, help="Only coadd this detector number")
    parser.add_argument("--obj", type=str, default=None,
                        help="Object name in lieu of extension, e.g if the spec2d files are named "
                             "'spec2d_J1234+5678_GNIRS_2017Mar31T085412.181.fits. then obj=J1234+5678")
    parser.add_argument("--show", default=False, action="store_true",
                        help="Show the reduction steps. Equivalent to the -s option when running pypeit.")
    parser.add_argument("--debug_offsets", default=False, action="store_true",
                        help="Show QA plots useful for debugging automatic offset determination")
    parser.add_argument("--peaks", default=False, action="store_true",
                        help="Show the peaks found by the object finding algorithm.")
    parser.add_argument("--basename", type=str, default=None,
                        help="Basename of files to save the parameters, spec1d, and spec2d")
    parser.add_argument('--samp_fact', default=1.0, type=float,
                        help="Make the wavelength grid finer (samp_fact > 1.0) or coarser (samp_fact < 1.0) by this sampling factor")
    parser.add_argument("--debug", default=False, action="store_true", help="show debug plots?")

    # TODO implement an option to only do certian slits
    #parser.add_argument("--only_slits", type=list, default=None, help="Only coadd the following slits")

    #parser.add_argument("--wave_method", type=str, default=None, help="Wavelength method for wavelength grid. If not set, code will use"
    #                                                                  "linear for Multislit and log10 for Echelle")
    #parser.add_argument("--std", default=False, action="store_true",
    #                    help="This is a standard star reduction.")


    return parser.parse_args() if options is None else parser.parse_args(options)


def main(args):
    """ Executes 2d coadding
    """
    msgs.warn('PATH =' + os.getcwd())
    # Load the file
    if args.file is not None:
        spectrograph, config_lines, spec2d_files = read_coadd2d_file(args.file)
        # Parameters
        # TODO: Shouldn't this reinstantiate the same parameters used in
        # the PypeIt run that extracted the objects?  Why are we not
        # just passing the pypeit file?
        # JFH: The reason is that the coadd2dfile may want different reduction parameters
        spectrograph_def_par = spectrograph.default_pypeit_par()
        parset = par.PypeItPar.from_cfg_lines(cfg_lines=spectrograph_def_par.to_config(),
                                                 merge_with=config_lines)
    elif args.obj is not None:
        # TODO: We should probably be reading the pypeit file and using those parameters here rather than using the
        # default parset.
        # TODO: This needs to define the science path
        spec2d_files = glob.glob('./Science/spec2d_*' + args.obj + '*')
        head0 = fits.getheader(spec2d_files[0])
        spectrograph_name = head0['SPECTROG']
        spectrograph = load_spectrograph(spectrograph_name)
        parset = spectrograph.default_pypeit_par()
    else:
        msgs.error('You must either input a coadd2d file with --file or an object name with --obj')

    # Update with configuration specific parameters (which requires science file) and initialize spectrograph
    spectrograph_cfg_lines = spectrograph.config_specific_par(spec2d_files[0]).to_config()
    parset = par.PypeItPar.from_cfg_lines(cfg_lines=spectrograph_cfg_lines, merge_with=parset.to_config())

    # If detector was passed as an argument override whatever was in the coadd2d_file
    if args.det is not None:
        msgs.info("Restricting reductions to detector={}".format(args.det))
        parset['rdx']['detnum'] = int(args.det)

    # Get headers (if possible) and base names
    spec1d_files = [files.replace('spec2d', 'spec1d') for files in spec2d_files]
    head1d = None
    for spec1d_file in spec1d_files:
        if os.path.isfile(spec1d_file):
            head1d = fits.getheader(spec1d_file)
            break
    if head1d is None:
        msgs.warn("No 1D spectra so am generating a dummy header for output")
        head1d = io.initialize_header()

    head2d = fits.getheader(spec2d_files[0])
    if args.basename is None:
        filename = os.path.basename(spec2d_files[0])
        basename = filename.split('_')[2]
    else:
        basename = args.basename

    # Write the par to disk
    par_outfile = basename+'_coadd2d.par'
    print("Writing the parameters to {}".format(par_outfile))
    parset.to_config(par_outfile)

    # Now run the coadds

    skysub_mode = head2d['SKYSUB']
    ir_redux = True if 'DIFF' in skysub_mode else False

    # Print status message
    msgs_string = 'Reducing target {:s}'.format(basename) + msgs.newline()
    msgs_string += 'Performing coadd of frames reduce with {:s} imaging'.format(skysub_mode)
    msgs_string += msgs.newline() + 'Combining frames in 2d coadd:' + msgs.newline()
    for file in spec2d_files:
        msgs_string += '{0:s}'.format(os.path.basename(file)) + msgs.newline()
    msgs.info(msgs_string)

    # TODO: This needs to be added to the parameter list for rdx
    redux_path = os.getcwd()
    master_dirname = os.path.basename(head2d['PYPMFDIR']) + '_coadd'
    master_dir = os.path.join(redux_path, master_dirname)

    # Make the new Master dir
    if not os.path.isdir(master_dir):
        msgs.info('Creating directory for Master output: {0}'.format(master_dir))
        os.makedirs(master_dir)

    # Instantiate the sci_dict
    sci_dict = OrderedDict()  # This needs to be ordered
    sci_dict['meta'] = {}
    sci_dict['meta']['vel_corr'] = 0.
    sci_dict['meta']['ir_redux'] = ir_redux

    # Find the detectors to reduce
    detectors = PypeIt.select_detectors(detnum=parset['rdx']['detnum'], ndet=spectrograph.ndet)
    if len(detectors) != spectrograph.ndet:
        msgs.warn('Not reducing detectors: {0}'.format(' '.join([str(d) for d in
        set(np.arange(spectrograph.ndet) + 1) - set(detectors)])))

    # Loop on detectors
    for det in detectors:
        msgs.info("Working on detector {0}".format(det))
        sci_dict[det] = {}

        # Instantiate Coadd2d
        coadd = coadd2d.CoAdd2D.get_instance(spec2d_files, spectrograph, parset, det=det,
                                             offsets=parset['coadd2d']['offsets'],
                                             weights=parset['coadd2d']['weights'],
                                             ir_redux=ir_redux,
                                             debug_offsets=args.debug_offsets, debug=args.debug,
                                             samp_fact=args.samp_fact)

        # Coadd the slits
        coadd_dict_list = coadd.coadd(only_slits=None) # TODO implement only_slits later
        # Create the psuedo images
        psuedo_dict = coadd.create_psuedo_image(coadd_dict_list)
        # Reduce
        msgs.info('Running the extraction')
        sci_dict[det]['sciimg'], sci_dict[det]['sciivar'], sci_dict[det]['skymodel'], sci_dict[det]['objmodel'], \
        sci_dict[det]['ivarmodel'], sci_dict[det]['outmask'], sci_dict[det]['specobjs'] = coadd.reduce(
            psuedo_dict, show = args.show, show_peaks = args.peaks)
        # Save psuedo image master files
        coadd.save_masters(master_dir)

    # Make the new Science dir
    # TODO: This needs to be defined by the user
    scipath = os.path.join(redux_path, 'Science_coadd')
    if not os.path.isdir(scipath):
        msgs.info('Creating directory for Science output: {0}'.format(scipath))
        os.makedirs(scipath)

    # Save the results
    save.save_all(sci_dict, coadd.stack_dict['master_key_dict'], master_dir, spectrograph, head1d,
                  head2d, scipath, basename)#, binning=coadd.binning)






#!/usr/bin/env python
#
# See top-level LICENSE file for Copyright information
#
# -*- coding: utf-8 -*-
"""
This script converts a LowRedux pixel flat into a PYPIT ready one
"""
import argparse

def parser(options=None):

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('lowrdx_file', type = str, default = None,
                        help = 'LowRedux Pixel Flat FITS file')
    parser.add_argument('new_file', type = str, default = None, help = 'PYPIT FITS file')

    args = parser.parse_args()

    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)
    return args


def main(args):
    try:
        from xastropy.xutils import xdebug as debugger
    except:
        import pdb as debugger

    from pypeit import arlris

    # Assume LRIS for now
    arlris.convert_lowredux_pixflat(args.lowrdx_file, args.new_file)




"""
Module to test spectrograph read functions
"""
import os

import pytest
import glob

from pkg_resources import resource_filename

from pypeit import spectrographs
from pypeit.core import procimg

from pypeit.tests.tstutils import dev_suite_required

# TODO: Add a test for Gemini GNIRS

@dev_suite_required
def test_keckdeimos():
    s = spectrographs.keck_deimos.KeckDEIMOSSpectrograph()
    example_file = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'keck_deimos',
                                '830G_L_8400', 'd0914_0002.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for Keck DEIMOS read.'
    det = 2
    #data, _ = s.load_raw_frame(example_file, det=det)
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    #
    bpm = s.bpm(example_file, det) #shape=shape) # filename=example_file)
    assert data.shape == (4096,2128)
    assert bpm.shape == (4096,2048)


@dev_suite_required
def test_kecklrisblue():
    s = spectrographs.keck_lris.KeckLRISBSpectrograph()
    example_file = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'keck_lris_blue',
                                'long_400_3400_d560', 'LB.20160109.14149.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for Keck LRIS blue read.'
    det = 2
    #data, _ = s.load_raw_frame(example_file, det=det)
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    #
    bpm = s.bpm(example_file, det) #shape=shape) # filename=example_file)
    assert data.shape == (2048,1154)
    assert bpm.shape == (2048,1024)


@dev_suite_required
def test_kecklrisred():
    s = spectrographs.keck_lris.KeckLRISRSpectrograph()
    example_file = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'keck_lris_red',
                                'long_600_7500_d560', 'LR.20160216.05529.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for Keck LRIS red read.'
    det = 1
    #data, _ = s.load_raw_frame(example_file, det=det)
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    #
    bpm = s.bpm(example_file, det)#, debug=True) #shape=shape) # filename=example_file)
    assert data.shape == (2068,1110)
    assert bpm.shape == (2048,1024)


@dev_suite_required
def test_kecknires():
    s = spectrographs.keck_nires.KeckNIRESSpectrograph()
    # TODO: Any Keck NIRES files to read?


@dev_suite_required
def test_kecknirspec():
    s = spectrographs.keck_nirspec.KeckNIRSPECSpectrograph()
    example_file = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'keck_nirspec',
                                'LOW_NIRSPEC-1', 'NS.20160414.02637.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for Keck NIRSPEC read.'
    #data, _ = s.load_raw_frame(example_file)
    det=1
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    bpm = s.bpm(example_file, det)
    assert data.shape == bpm.shape, 'Image and BPM have different shapes!'


def test_shanekastblue():
    s = spectrographs.shane_kast.ShaneKastBlueSpectrograph()
    example_file = os.path.join(resource_filename('pypeit', 'tests'), 'files',
                                'b1.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for Shane Kast blue read.'
    det=1
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    bpm = s.bpm(example_file, det)
    assert data.shape == (350, 2112)
    assert bpm.shape == (2048,350)


@dev_suite_required
def test_shanekastredret():
    s = spectrographs.shane_kast.ShaneKastRedRetSpectrograph()
    example_file = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'shane_kast_red',
                                '600_7500_d55_ret', 'r112.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for Shane Kast red read.'
    det = 1
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    bpm = s.bpm(example_file, det)
    assert data.shape == (250, 1232)
    assert bpm.shape == (1200, 250)


def test_shanekastred():
    spectrographs.shane_kast.ShaneKastRedSpectrograph()
    # TODO: Any Shane Kast Red files to read?


def test_tngdolores():
    s = spectrographs.tng_dolores.TNGDoloresSpectrograph()
    # TODO: Any TNG Dolores files to read?


@dev_suite_required
def test_vltxshooteruvb():
    s = spectrographs.vlt_xshooter.VLTXShooterUVBSpectrograph()
    example_file = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'vlt_xshooter',
                                'UVB_1x1', 'XSHOO.2010-04-28T05:34:32.723.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for VLT Xshooter UVB read.'
    det = 1
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    bpm = s.bpm(example_file, det)
    assert data.shape == (3000, 2144)
    assert bpm.shape == (3000, 2048)


@dev_suite_required
def test_vltxshootervis():
    s = spectrographs.vlt_xshooter.VLTXShooterVISSpectrograph()
    example_file = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'vlt_xshooter',
                                'VIS_1x1', 'XSHOO.2010-04-28T05:34:37.853.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for VLT Xshooter VIS read.'
    det = 1
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    bpm = s.bpm(example_file, det)
    assert data.shape == (4000, 2106)
    assert bpm.shape == (4000, 2048)


@dev_suite_required
def test_vltxshooternir():
    s = spectrographs.vlt_xshooter.VLTXShooterNIRSpectrograph()
    example_file = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'vlt_xshooter',
                                'NIR', 'XSHOO.2016-08-02T08:45:49.494.fits.gz')
    assert os.path.isfile(example_file), 'Could not find example file for VLT Xshooter NIR read.'
    det = 1
    data, hdu, exptime, rawdatasec_img, oscansec_img = s.get_rawimage(example_file, det)
    bpm = s.bpm(example_file, det)
    assert data.shape == (1100,2048)
    assert bpm.shape == (2045, 1097)






#!/usr/bin/env python
#
# See top-level LICENSE file for Copyright information
#
# -*- coding: utf-8 -*-
"""
This script runs PypeIt
"""
import argparse

from pypeit import msgs

import warnings

def parser(options=None):

    parser = argparse.ArgumentParser(description=msgs.usage('PypeIt'),
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('pypeit_file', type=str,
                        help='PypeIt reduction file (must have .pypeit extension)')
    parser.add_argument('-v', '--verbosity', type=int, default=2,
                        help='Verbosity level between 0 [none] and 2 [all]')
    # JFH TODO Are the -t and -r keyword still valid given that run_pypeit no longer runs setup?
    parser.add_argument('-t', '--hdrframetype', default=False, action='store_true',
                        help='Use file headers and the instument-specific keywords to determine'
                             'the type of each frame')
    parser.add_argument('-r', '--sort_dir', default=None,
                        help='Directory used to store the sorted files.  Default is to omit '
                             'writing these files.')
    parser.add_argument('-m', '--use_masters', default=False, action='store_true',
                        help='Load previously generated MasterFrames')
    parser.add_argument('-s', '--show', default=False, action='store_true',
                        help='Show reduction steps via plots (which will block further execution until clicked on) '
                             'and outputs to ginga. Requires remote control ginga session via "ginga --modules=RC &"')
    # JFH Should the default now be true with the new definition.
    parser.add_argument('-o', '--overwrite', default=False, action='store_true',
                        help='Overwrite any existing files/directories')
    group = parser.add_mutually_exclusive_group()
#    group.add_argument('-p', '--prep_setup', default=False, action='store_true',
#                       help='Run pypeit to prepare the setup only')
#    group.add_argument('-c', '--calcheck', default=False, action='store_true',
#                       help='Run pypeit only as a check on the calibrations')
    group.add_argument('-d', '--detector', default=None, help='Detector to limit reductions on.  If the output files exist and -o is used, the outputs for the input detector will be replaced.')

#    parser.add_argument('-q', '--quick', default=False, help='Quick reduction',
#                        action='store_true')
#    parser.add_argument('-c', '--cpus', default=False, action='store_true',
#                         help='Number of CPUs for parallel processing')
#    parser.print_help()

    if options is None:
        pargs = parser.parse_args()
    else:
        pargs = parser.parse_args(options)
    return pargs


def main(args):

    import os

    from pypeit import pypeit

    # Initiate logging for bugs and command line help
    # These messages will not be saved to a log file
    # Set the default variables
    qck = False
    cpu = 1
    #vrb = 2

    # Load options from command line
    splitnm = os.path.splitext(args.pypeit_file)
    if splitnm[1] != '.pypeit':
        msgs.error("Bad extension for PypeIt reduction file."+msgs.newline()+".pypeit is required")
    logname = splitnm[0] + ".log"

    # Instantiate the main pipeline reduction object
    pypeIt = pypeit.PypeIt(args.pypeit_file, verbosity=args.verbosity,
                           reuse_masters=args.use_masters, overwrite=args.overwrite,
                           logname=logname, show=args.show)

    # JFH I don't see why this is an optional argument here. We could allow the user to modify an infinite number of parameters
    # from the command line? Why do we have the PypeIt file then? This detector can be set in the pypeit file.
    # Detector?
    if args.detector is not None:
        msgs.info("Restricting reductions to detector={}".format(args.detector))
        pypeIt.par['rdx']['detnum'] = int(args.detector)

    pypeIt.reduce_all()
    msgs.info('Data reduction complete')
    # QA HTML
    msgs.info('Generating QA HTML')
    pypeIt.build_qa()

    return 0

